/*
 This file is part of Subsonic.

 Subsonic is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Subsonic is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Subsonic.  If not, see <http://www.gnu.org/licenses/>.

 Copyright 2009 (C) Sindre Mehus
 */
package net.sourceforge.subsonic.androidapp.service;

import android.content.Context;
import android.graphics.Bitmap;
import android.util.Log;
import net.sourceforge.subsonic.androidapp.domain.Indexes;
import net.sourceforge.subsonic.androidapp.domain.MusicDirectory;
import net.sourceforge.subsonic.androidapp.domain.Playlist;
import net.sourceforge.subsonic.androidapp.domain.Version;
import net.sourceforge.subsonic.androidapp.domain.SearchResult;
import net.sourceforge.subsonic.androidapp.domain.SearchCritera;
import net.sourceforge.subsonic.androidapp.util.CancellableTask;
import net.sourceforge.subsonic.androidapp.util.LRUCache;
import net.sourceforge.subsonic.androidapp.util.ProgressListener;
import net.sourceforge.subsonic.androidapp.util.TimeLimitedCache;
import net.sourceforge.subsonic.androidapp.util.Util;
import net.sourceforge.subsonic.androidapp.util.FileUtil;
import net.sourceforge.subsonic.androidapp.util.Exceptions;
import org.apache.http.HttpResponse;

import java.util.Collection;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * @author Sindre Mehus
 */
public class CachedMusicService implements MusicService {

    private static final String TAG = CachedMusicService.class.getSimpleName();

    private static final int MUSIC_DIR_CACHE_SIZE = 20;
    private static final int TTL_MUSIC_DIR = 5 * 60; // Five minutes
    private static final long DELAY_MS_PER_SONG = 2 * 60 * 1000;  // Two minutes (conservative)

    private final MusicService musicService;
    private final LRUCache<String, TimeLimitedCache<MusicDirectory>> cachedMusicDirectories;
    private final TimeLimitedCache<Boolean> cachedLicenseValid = new TimeLimitedCache<Boolean>(120, TimeUnit.SECONDS);
    private final TimeLimitedCache<Indexes> cachedIndexes = new TimeLimitedCache<Indexes>(60 * 60, TimeUnit.SECONDS);
    private final TimeLimitedCache<List<Playlist>> cachedPlaylists = new TimeLimitedCache<List<Playlist>>(60, TimeUnit.SECONDS);
    private String restUrl;

    public CachedMusicService(MusicService musicService) {
        this.musicService = musicService;
        cachedMusicDirectories = new LRUCache<String, TimeLimitedCache<MusicDirectory>>(MUSIC_DIR_CACHE_SIZE);
    }

    @Override
    public void ping(Context context, ProgressListener progressListener) throws Exception {
        checkSettingsChanged(context);
        musicService.ping(context, progressListener);
    }

    @Override
    public boolean isLicenseValid(Context context, ProgressListener progressListener) throws Exception {
        checkSettingsChanged(context);
        Boolean result = cachedLicenseValid.get();
        if (result == null) {
            result = musicService.isLicenseValid(context, progressListener);
            cachedLicenseValid.set(result);
        }
        return result;
    }

    @Override
    public Indexes getFastIndexes(Context context) throws Exception {
        return this.musicService.getFastIndexes(context);
    }

    @Override
    public Indexes getIndexes(boolean refresh, Context context, ProgressListener progressListener, long lastModified) throws Exception {
        checkSettingsChanged(context);
        Indexes result = refresh ? null : cachedIndexes.get();
        if (result == null) {
            result = musicService.getIndexes(refresh, context, progressListener, lastModified);
            cachedIndexes.set(result);
        }
        return result;
    }

    @Override
    public MusicDirectory getMusicDirectory(String id, Context context, ProgressListener progressListener) throws Exception {
        checkSettingsChanged(context);
        TimeLimitedCache<MusicDirectory> cache = cachedMusicDirectories.get(id);
        MusicDirectory dir = cache == null ? null : cache.get();
        if (dir == null) {
            dir = musicService.getMusicDirectory(id, context, progressListener);
            cache = new TimeLimitedCache<MusicDirectory>(TTL_MUSIC_DIR, TimeUnit.SECONDS);
            cache.set(dir);
            cachedMusicDirectories.put(id, cache);
        }
        return dir;
    }

    @Override
    public SearchResult search(SearchCritera criteria, Context context, ProgressListener progressListener) throws Exception {
        return musicService.search(criteria, context, progressListener);
    }

    @Override
    public MusicDirectory getPlaylist(String id, Context context, ProgressListener progressListener) throws Exception {
        return musicService.getPlaylist(id, context, progressListener);
    }

    @Override
    public List<Playlist> getPlaylists(Context context, ProgressListener progressListener) throws Exception {
        checkSettingsChanged(context);
        List<Playlist> result = cachedPlaylists.get();
        if (result == null) {
            result = musicService.getPlaylists(context, progressListener);
            cachedPlaylists.set(result);
        }
        return result;
    }

    @Override
    public void createPlaylist(String id, String name, List<MusicDirectory.Entry> entries, Context context, ProgressListener progressListener) throws Exception {
        musicService.createPlaylist(id, name, entries, context, progressListener);
    }
    
    @Override
    public void createPlaylist(String id, String name, Collection<MusicDirectory.Entry> entries, Context context, ProgressListener progressListener) throws Exception {
    	musicService.createPlaylist(id, name, entries, context, progressListener);
    }

    @Override
    public MusicDirectory getAlbumList(String type, int size, int offset, Context context, ProgressListener progressListener) throws Exception {
        return musicService.getAlbumList(type, size, offset, context, progressListener);
    }

    void updateGetRandomSongs(final int size, final Context context, final ProgressListener progressListener, final long sleepDuration) throws Exception {

        Thread t = new Thread("Refresh Random-Songs Cache") {
            @Override
            public void run() {
                Log.i(TAG, "Getting new list of random songs in " + (sleepDuration/1000) + " seconds.");
                if (sleepDuration != 0L) {
                    Log.i(TAG, "Not requesting new list of songs right away.");
                    try {
                        sleep(sleepDuration);
                    } catch (InterruptedException ex) {
                        Log.i(TAG, "Something interrupted us. Aborting.  Not refreshing random-songs.");
                        return;
                    }
                }
                try {
                    MusicDirectory fresh = musicService.getRandomSongs(size, context, progressListener);
                    FileUtil.serialize(context, fresh, "getRandomSongs");
                } catch (Exception ex) {
                    Log.w(TAG, "Not caching next random songs.", ex);
                }
            }
        };
        t.start();
    }

    @Override
    public MusicDirectory getRandomSongs(final int size, final Context context, final ProgressListener progressListener) throws Exception {
        MusicDirectory cached = FileUtil.deserialize(context, "getRandomSongs");
        if ((cached == null) || (cached.getChildren().size() == 0)) {
            Log.d(TAG, "had no cache of random songs. Creating.");
            // Schedule refresh.  Returning null for now.
            updateGetRandomSongs(size, context, progressListener, 0L);
        } else {
            Log.d(TAG, "loaded cached random songs.  Found " + cached.getChildren().size());
            // We send a list of cached back.  Let's update the cache for next
            // time, if we're still running.
            updateGetRandomSongs(size, context, progressListener, size * DELAY_MS_PER_SONG);
        }
        return cached;
    }


    @Override
    public Bitmap getCoverArt(Context context, MusicDirectory.Entry entry, int size, ProgressListener progressListener) throws Exception {
        return musicService.getCoverArt(context, entry, size, progressListener);
    }

    @Override
    public HttpResponse getDownloadInputStream(Context context, MusicDirectory.Entry song, long offset, int maxBitrate, CancellableTask task) throws Exception {
        return musicService.getDownloadInputStream(context, song, offset, maxBitrate, task);
    }

    @Override
    public Version getLocalVersion(Context context) throws Exception {
        return musicService.getLocalVersion(context);
    }

    @Override
    public Version getLatestVersion(Context context, ProgressListener progressListener) throws Exception {
        return musicService.getLatestVersion(context, progressListener);
    }

    private void checkSettingsChanged(Context context) throws Exceptions.PrerequisiteNotSatisfied {
        String newUrl = Util.getRestUrl(context, null);
        if (!Util.equals(newUrl, restUrl)) {
            cachedMusicDirectories.clear();
            cachedLicenseValid.clear();
            cachedIndexes.clear();
            cachedPlaylists.clear();
            restUrl = newUrl;
        }
    }
}
