package net.sourceforge.subsonic.androidapp.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import android.os.Handler;
import android.content.Context;
import android.util.Log;
import net.sourceforge.subsonic.androidapp.service.DownloadFile;
import net.sourceforge.subsonic.androidapp.service.DownloadService;

import net.sourceforge.subsonic.u1m.R;

/**
 * @author Sindre Mehus
 * @version $Id$
 */
public class CacheCleaner {

    private static final String TAG = CacheCleaner.class.getSimpleName();

    private final Context context;
    private final DownloadService downloadService;
    private final Handler notificationChannel;

    public CacheCleaner(Context context, DownloadService downloadService, Handler notificationChannel) {
        this.context = context;
        this.downloadService = downloadService;
        this.notificationChannel = notificationChannel;
    }

    public void clean() {
        Log.i(TAG, "Starting cache cleaning.");
        try {

            List<File> files = new ArrayList<File>();
            List<File> dirs = new ArrayList<File>();

            findCandidatesForDeletion(FileUtil.getMusicDirectory(), files, dirs);
            sortByAscendingModificationTime(files);

            Set<File> undeletable = findUndeletableFiles();

            deleteFiles(files, undeletable);
            deleteEmptyDirs(dirs, undeletable);
            Log.i(TAG, "Completed cache cleaning.");

        } catch (RuntimeException x) {
            Log.e(TAG, "Error in cache cleaning.", x);
            throw x;
        }
    }

    private void deleteEmptyDirs(List<File> dirs, Set<File> undeletable) {
        for (File dir : dirs) {
            if (undeletable.contains(dir)) {
                continue;
            }

            // Delete album art if it's the only remaining file.
            File[] children = dir.listFiles();
            if (children.length == 1 && Constants.ALBUM_ART_FILE.equals(children[0].getName())) {
                Util.delete(children[0]);
                children = dir.listFiles();
            }

            // Delete empty directory.
            if (children.length == 0) {
                Util.delete(dir);
            }
        }
    }

    private void deleteFiles(List<File> files, Set<File> undeletable) {
        Integer cacheSizeMBytes = Util.getCacheSizeMB(context);
        final long cacheSizeBytes = (cacheSizeMBytes == null) ? (-1) : (cacheSizeMBytes * 1024L * 1024L);

        final long bytesUsed = FileUtil.getCacheUsedB(files);

        Log.i(TAG, "Cache size limit: " + Util.formatBytes(cacheSizeBytes));
        Log.i(TAG, "Cache used before: " + bytesUsed);

        // Cleaner is run every 5min.  Assume max 4MB per min. in playing.
        final long bytesReservedSpace = 5 * 4 * 1024 * 1024 ;  // For upcoming downloads.

        // How much we need to remove to limit by cache setting.
        long bytesNeedRemoved = 0;
        if (cacheSizeBytes > 0)
            bytesNeedRemoved = Math.max(bytesNeedRemoved, (bytesReservedSpace + bytesUsed) - cacheSizeBytes);

        // How much we need to remove to limit by actual filesystem capacity.  Take max().
        try {
            final long totalAvailableBytes = FileUtil.getFilesystemAvailableMB() * 1024 * 1024;
            final long remainingBytesIfRemovedEntireCache = totalAvailableBytes + bytesUsed;

            // Warn user that the cache-size setting is too high for this disk.
            if (cacheSizeBytes > remainingBytesIfRemovedEntireCache) {
                notificationChannel.post(new Runnable() {
                        @Override
                        public void run() {
                            Util.toastImportant(
                                context.getApplicationContext(), 
                                context.getString(
                                    R.string.cachecleaner_cache_gt_free, 
                                    Util.formatBytes(cacheSizeBytes),
                                    Util.formatBytes(cacheSizeBytes - remainingBytesIfRemovedEntireCache)));
                        }
                    });
                Log.w(TAG, "Disk space insufficient to hold " + Util.formatBytes(cacheSizeBytes) + " cache.  Could only have " + Util.formatBytes(remainingBytesIfRemovedEntireCache));
            }

            if (totalAvailableBytes < bytesReservedSpace) {
                if (cacheSizeBytes > 0) {
                    notificationChannel.post(new Runnable() {
                            @Override
                            public void run() {
                                Util.toastImportant(context.getApplicationContext(), R.string.cachecleaner_disk_full);
                            }
                        });
                }
                bytesNeedRemoved = Math.max(bytesNeedRemoved, bytesReservedSpace);
                Log.w(TAG, "Disk is too full.  Want to reserve " + Util.formatBytes(bytesReservedSpace) + " but I only have " + Util.formatBytes(totalAvailableBytes));
            }

        } catch (IllegalArgumentException e) {
            Log.w(TAG, "Couldln't check viability of cache size.", e);
        } catch (java.io.IOException e) {
            Log.w(TAG, "Couldln't check viability of cache size.", e);
        }

        int bytesRemoved = 0;
        for (File file : files) {
            if (bytesRemoved < bytesNeedRemoved || file.getName().endsWith(".partial")) {
                if (!undeletable.contains(file)) {
                    long size = file.length();
                    if (Util.delete(file)) {
                        bytesRemoved += size;
                    }
                } else {
                    Log.d(TAG, "Refusing to delete undeletable file " + file.toString());
                }
            }
        }

        if (bytesNeedRemoved > bytesRemoved) {
            //Util.toast(context.getApplicationContext(), "Couldn't make enough free space to play.");
            Log.w(TAG, "After deletion run, still want another " + Util.formatBytes(bytesNeedRemoved - bytesRemoved) + " free.");
        }

        Log.i(TAG, "Cache used after: " + Util.formatBytes(bytesUsed - bytesRemoved));
    }

    private void findCandidatesForDeletion(File file, List<File> files, List<File> dirs) {
        if (file.isFile()) {
            String name = file.getName();
            boolean isCacheFile = name.endsWith(".partial") || name.endsWith(".complete") || name.contains(".complete.");
            if (isCacheFile) {
                files.add(file);
            }
        } else if (file.isDirectory()) {
            // Depth-first
            for (File child : FileUtil.listFiles(file)) {
                findCandidatesForDeletion(child, files, dirs);
            }
            dirs.add(file);
        } else {
            Log.d(TAG, "Not considering for deletion nonfile, nondir entity " + file.toString());
        }
    }

    private void sortByAscendingModificationTime(List<File> files) {
        Collections.sort(files, new Comparator<File>() {
            @Override
            public int compare(File a, File b) {
                if (a.lastModified() < b.lastModified()) {
                    return -1;
                }
                if (a.lastModified() > b.lastModified()) {
                    return 1;
                }
                return 0;
            }
        });
    }

    private Set<File> findUndeletableFiles() {
        Set<File> undeletable = new HashSet<File>(5);

        for (DownloadFile downloadFile : downloadService.getDownloads()) {
            undeletable.add(downloadFile.getPartialFile());
            undeletable.add(downloadFile.getCompleteFile());
        }

        undeletable.add(FileUtil.getMusicDirectory());
        return undeletable;
    }
}
