/*
 This file is part of Subsonic.

 Subsonic is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Subsonic is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Subsonic.  If not, see <http://www.gnu.org/licenses/>.

 Copyright 2009 (C) Sindre Mehus
 */
package net.sourceforge.subsonic.androidapp.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Arrays;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.List;
import android.os.StatFs;

import android.content.Context;
import android.os.Environment;
import android.util.Log;
import net.sourceforge.subsonic.androidapp.domain.MusicDirectory;

/**
 * @author Sindre Mehus
 */
public class FileUtil {

    private static final String TAG = FileUtil.class.getSimpleName();
    private static final String[] FILE_SYSTEM_UNSAFE = {"/", "\\", "..", ":", "\"", "?", "*", "<", ">"};
    private static final List<String> MUSIC_FILE_EXTENSIONS = Arrays.asList("mp3", "ogg", "aac", "flac", "m4a", "wav", "wma");
    private static final File MUSIC_DIR = createDirectory("music");

    public static long getFilesystemAvailableMB() throws java.io.IOException {
        StatFs st = new StatFs(getMusicDirectory().getCanonicalPath());
        long blocksAvailable = (long) st.getAvailableBlocks();
        long blockSizeBytes = (long) st.getBlockSize();
        return blocksAvailable * blockSizeBytes / 1024L / 1024L;
    }

    public static long getCacheUsedB(List<File> files) {
        long bytesUsed = 0L;
        for (File file : files) {
            bytesUsed += file.length();
        }
        return bytesUsed;
    }

    public static File getSongFile(MusicDirectory.Entry song) {
        File dir = getAlbumDirectory(song);

        StringBuilder fileName = new StringBuilder();
        Integer track = song.getTrack();
        if (track != null) {
            if (track < 10) {
                fileName.append("0");
            }
            fileName.append(track).append("-");
        }

        fileName.append(fileSystemSafe(song.getTitle()));

		if ("Unknown artist".equals(song.getArtist()) &&
		    "Unknown album".equals(song.getAlbum()) && 
		    "Unknown title".equals(song.getTitle())) {

			fileName.append(" (id");
			fileName.append(song.getId().replace(':', '-'));
			fileName.append(")");
		}

		fileName.append(".");

        if (song.getTranscodedSuffix() != null) {
            fileName.append(song.getTranscodedSuffix());
        } else {
            fileName.append(song.getSuffix());
        }

		File ret = new File(dir, fileName.toString());
		Log.i(TAG, "File path is " + ret.toString());

        return new File(dir, fileName.toString());
    }

    public static File getAlbumArtFile(MusicDirectory.Entry entry) {

        File dir = getAlbumDirectory(entry);
        File oldSchemeAlbumArtFile = new File(dir, Constants.ALBUM_ART_FILE);
        if (oldSchemeAlbumArtFile.exists()) {
            oldSchemeAlbumArtFile.delete();  // Declutter.
        }

        File externalStorage = Environment.getExternalStorageDirectory();
        File applicationDirectory = new File(externalStorage, "Android/data/com.ubuntu.one.music/files");
        File albumArtDirectory = new File(applicationDirectory, "albumart");
        File noMediaFlag = new File(albumArtDirectory, ".nomedia");

        if (! albumArtDirectory.exists()) {
            albumArtDirectory.mkdirs();
        }
        if (! noMediaFlag.exists()) {
            try {
                noMediaFlag.createNewFile();
            } catch (java.io.IOException ex) {
                Log.i(TAG, "Can't make media-scanner excluder " + noMediaFlag);
            }
        }
        if (entry.isDirectory()) {
            return new File(albumArtDirectory, fileSystemSafe(entry.getArtist()) + "," + fileSystemSafe(entry.getTitle()));
        } else {
            return new File(albumArtDirectory, fileSystemSafe(entry.getArtist()) + "," + fileSystemSafe(entry.getAlbum()));
        }
    }

    private static File getAlbumDirectory(MusicDirectory.Entry song) {
        File dir;
        // Fallback position:  Construct a default
        File artistDir = new File(MUSIC_DIR.getPath(), fileSystemSafe(song.getArtist()));
        dir = new File(artistDir, fileSystemSafe(song.getAlbum()));

        // Now see if there's a better answer.
        if (song.getPath() != null) {
            File f = new File(song.getPath());
            // Use path-groking Java library to decide if we have a parent or not.
            if (f.getParent() != null)
                dir = new File(MUSIC_DIR.getPath(), f.getParent());
        }
        return dir;
    }

    public static void createDirectoryForParent(File file) {
        File dir = file.getParentFile();
        if (!dir.exists()) {
            if (!dir.mkdirs()) {
                Log.e(TAG, "Failed to create directory " + dir);
            }
        }
    }

    private static File createDirectory(String name) {
        File subsonicDir = new File(Environment.getExternalStorageDirectory(), "subsonic");
        File dir = new File(subsonicDir, name);
        if (!dir.exists() && !dir.mkdirs()) {
            Log.e(TAG, "Failed to create " + name);
        }
        return dir;
    }

    public static File getMusicDirectory() {
        return MUSIC_DIR;
    }

    /**
     * Makes a given filename safe by replacing special characters like slashes ("/" and "\")
     * with dashes ("-").
     *
     * @param filename The filename in question.
     * @return The filename with special characters replaced by underscores.
     */
    private static String fileSystemSafe(String filename) {
        if (filename == null || filename.trim().length() == 0) {
            return "unnamed";
        }

        for (String s : FILE_SYSTEM_UNSAFE) {
            filename = filename.replace(s, "-");
        }
        return filename;
    }

    /**
     * Similar to {@link File#listFiles()}, but returns a sorted set.
     * Never returns {@code null}, instead a warning is logged, and an empty set is returned.
     */
    public static SortedSet<File> listFiles(File dir) {
        File[] files = dir.listFiles();
        if (files == null) {
            Log.w(TAG, "Failed to list children for " + dir.getPath());
            return new TreeSet<File>();
        }

        return new TreeSet<File>(Arrays.asList(files));
    }

    public static SortedSet<File> listMusicFiles(File dir) {
        SortedSet<File> files = listFiles(dir);
        Iterator<File> iterator = files.iterator();
        while (iterator.hasNext()) {
            File file = iterator.next();
            if (!file.isDirectory() && !isMusicFile(file)) {
                iterator.remove();
            }
        }
        return files;
    }

    private static boolean isMusicFile(File file) {
        String extension = getExtension(file.getName());
        return MUSIC_FILE_EXTENSIONS.contains(extension);
    }

    /**
     * Returns the extension (the substring after the last dot) of the given file. The dot
     * is not included in the returned extension.
     *
     * @param name The filename in question.
     * @return The extension, or an empty string if no extension is found.
     */
    public static String getExtension(String name) {
        int index = name.lastIndexOf('.');
        return index == -1 ? "" : name.substring(index + 1);
    }

    /**
     * Returns the base name (the substring before the last dot) of the given file. The dot
     * is not included in the returned basename.
     *
     * @param name The filename in question.
     * @return The base name, or an empty string if no basename is found.
     */
    public static String getBaseName(String name) {
        int index = name.lastIndexOf('.');
        return index == -1 ? name : name.substring(0, index);
    }

    public static <T> boolean serialize(Context context, T obj, String fileName) {
        File file = new File(context.getCacheDir(), fileName);
        ObjectOutputStream out = null;
        try {
            out = new ObjectOutputStream(new FileOutputStream(file));
            out.writeObject(obj);
            Log.i(TAG, "Serialized object to " + file);
            return true;
        } catch (Throwable x) {
            Log.w(TAG, "Failed to serialize object to " + file);
            return false;
        } finally {
            Util.close(out);
        }
    }

    @SuppressWarnings("unchecked")
    public static <T> T deserialize(Context context, String fileName) {
        File file = new File(context.getCacheDir(), fileName);
        if (!file.exists() || !file.isFile()) {
            Log.i(TAG, "no such file " + file);
            return null;
        }

        ObjectInputStream in = null;
        try {
            in = new ObjectInputStream(new FileInputStream(file));
            T result = (T) in.readObject();
            Log.i(TAG, "Deserialized object from " + file);
            return result;
        } catch (java.io.InvalidClassException x) {
            Log.w(TAG, "Failed to deserialize object from " + file + ".  Deleting.", x);
            file.delete();
            return null;
        } catch (Throwable x) {
            Log.w(TAG, "Failed to deserialize object from " + file, x);
            return null;
        } finally {
            Util.close(in);
        }
    }
}
